import numpy as np
import pylab as pl
from scipy.optimize import  fmin_l_bfgs_b
import string


def xcurve_5(x,x0,x1,x2,a,b):
    '''Solution to the differential equation
       d/dt rho = (x2 - rho)*(a+b*(rho-x1)) if rho in [x1,x2] and 0.0 else;
       with initial condition rho(x0) = 1/2*(y0+y1); 
       this curve goes from x1 to x2;
       note the point x0 is not the intial condition rho(x0) = x1 since this is ill-defined for a=0 (s-curve).
       the special cases are exp-curve (b=0.0) and s-curve (a=0.0)
       
       if a>0, there is a finite time t0: rho(t0)=x1, such that
       rho(t) = x1 for t<t0 and rho(t) = xcurve for t>=t0
       if a==0, rho(t-->-infty) --> x1 <--> t0 =infty
       
       In: x ... array for times at which to compute curve
           x0 ... parameter where rho(x0)=1/2*(y0+y1) 'half the change'
           x1 ... lower asymptotic
           x2 ... upper asymptotic
           a ... external influence
           b ... internal influence
       
    '''
    f_tmp = np.exp((a+b*(x2-x1))*(x-x0))
    if a>0:
        t0 = (x0*(a+b*(x2-x1)) - np.log((2.0*a+b*(x2-x1))/a))/(a+b*(x2-x1))
        y = (x<t0)*x1 + (x>=t0)*(-(a-b*x1)*(x2-x1) + x2*(2.0*a+b*(x2-x1))*f_tmp )/( b*(x2-x1) + (2.0*a+b*(x2-x1))*f_tmp )
    else:
        y = ( -(a-b*x1)*(x2-x1) + x2*(2.0*a+b*(x2-x1))*f_tmp )/( b*(x2-x1) + (2.0*a+b*(x2-x1))*f_tmp )
    return y
    
def fit_mixed(x,y,sigma,theta_0):
    '''fits the mixed model to data (x,y,sigma) with initial parameter theta_0 numerically (local minimum!).
       returns: [estimated parameters, chi_squared of best fit]
    '''
    bnds = [(None,None),(0.0,1.0),(0.0,1.0),(0.0,None),(0.0,None)]
    if sigma == None:
        sigma = np.ones(len(x))
    result_tmp = fmin_l_bfgs_b(ls_fit_mixed,theta_0,args=(x,y,sigma),bounds=bnds, approx_grad=True,disp=0,iprint=-1)
    chi2 = 0
    if result_tmp[2]['warnflag']==0:
        result = result_tmp[0]
        chi2 = ls_fit_mixed(result[:5],x,y,sigma)
    else:
        chi2 = 0
        result = [0,0,0,0,0]
        print 'Optimization did not converge - try different initial conditions for the parameters!'
    return [result,chi2]

def ls_fit_mixed(theta,x,y,sigma):
    '''calculates the chi_squared function for the mixed model with parameters theta compared to data (x,y,sigma).
    '''
    x0 = theta[0]
    x1 = theta[1]
    x2 = theta[2]
    a = theta[3]
    b = theta[4]
    Delta = np.sum(((y - xcurve_5(x,x0,x1,x2,a,b))/sigma)**2)
    return Delta
    
## import the data and create timeseries: t, rho, sigma
f = open('VerbsRegularization_en_abide.csv','r')
header = f.readline()
x = f.readlines()
f.close()
t = np.array([int(string.split(h)[0]) for h in x]) # time
rho = np.array([float(string.split(h)[1]) for h in x]) # rho
sigma_s = np.array([float(string.split(h)[2]) for h in x]) # error from finite sampling
sigma_0 = np.sqrt(0.002) # error from external sources
sigma = np.sqrt(sigma_s**2 + sigma_0**2  ) # combine sigma_0 and sigma_s

## fit the mixed model to the timeseries
# initial conditions for parameters in the numerical optimizations
# check different values in case of local minima in the Least-Square-minimization, e.g. setting random initial conditions and making amny runs
theta_0 = [0.5*(t[-1]+t[0]),0.1,0.9,0.01,0.1] 

# in this example, we formulate the mixed model only for increasing timeseries in rho.
# if the timeseries is decreasing we fit the transformed timeseries rho' = 1-rho (which is increasing)
# this is checked by comparing the average of the first 10 versus the last 10 points in the timeseries
bool_incr = True
if np.sum(rho[:10])>np.sum(rho[-10:]):
    rho = 1.0 - rho
    bool_incr = False

# fit the timeseries and extract the parameters
fit_result = fit_mixed(t,rho,sigma,theta_0) 
theta = fit_result[0] # estimated parameters [t0,y0,y1,a,b] from best fit
chi2 = fit_result[1] # chi_squared of the best fit

## plot the time series and the data
# timeseries for the mixed model with parameters of the best fit
t_fit = np.linspace(t[0],t[-1],1000,endpoint=True)
rho_fit = xcurve_5(t_fit,theta[0],theta[1],theta[2],theta[3],theta[4])

# if the original timeseries rho is decreasing, transform back to rho = 1 - rho'
if bool_incr == False:
    rho_fit = 1.0-rho_fit
    rho = 1.0-rho

pl.figure()
pl.xlabel('year')
pl.ylabel(r'$\rho$')
pl.plot(t,rho,marker='o',color='blue',lw=0,label='data',ms=3)
pl.plot(t_fit,rho_fit,lw=3,color='red',label='fit')
pl.legend(loc = 'upper left')
pl.ylim(0,1)
pl.show()
